/* -*- tab-width: 4 -*- */

/* Environments, define, set!, etc.
 */

#include "elkkernel.h"

#define Env_To_List(env, list) SET((list), T_Pair, POINTER(env))
#define List_To_Env(list, env) SET((env), T_Environment, POINTER(list))

ELKObject The_Environment, Global_Environment;

ELKObject General_Define(ELKObject argl, ELKObject sym);
void Memoize_Frames (ELKObject thiso, ELKObject last);
void Forget_Frame (ELKObject frame);
void Set_Name (ELKObject var, ELKObject val);
ELKObject Define_Procedure (ELKObject form, ELKObject body, ELKObject sym);

void Init_Env (void) {
	List_To_Env (Cons (Null, Null), Global_Environment);
	The_Environment = Global_Environment;
	Global_GC_Link (Global_Environment);
	Global_GC_Link (The_Environment);
}

ELKObject P_Environment_To_List (ELKObject env) {
	ELKObject e;

	Check_Type (env, T_Environment);
	Env_To_List (env, e);
	return Copy_List (e);
}

ELKObject P_Environmentp (ELKObject x) {
	return TYPE(x) == T_Environment ? Trueob : Falseob;
}

void Push_Frame (ELKObject frame) {
	ELKObject e;

	Memoize_Frame (frame);
	Env_To_List (The_Environment, e);
	List_To_Env (Cons (frame, e), The_Environment);
}

void Pop_Frame (void) {
	ELKObject e;

	Env_To_List (The_Environment, e);
	List_To_Env (Cdr (e), The_Environment);
	Forget_Frame (Car (e));
}

void Switch_Environment (ELKObject to) {
	ELKObject old, newo, n;

	if (EQ(The_Environment,to))
		return;
	Env_To_List (The_Environment, old);
	Env_To_List (to, newo);
	for ( ; !Nullp (old); old = Cdr (old)) {
		for (n = newo; !Nullp (n) && !EQ(n,old);
				n = Cdr (n))
			;
		if (EQ(n,old))
			break;
		Forget_Frame (Car (old));
	}
	Memoize_Frames (newo, n);
	The_Environment = to;
}

void Memoize_Frames (ELKObject thiso, ELKObject last) {
	if (Nullp (thiso) || EQ(thiso,last))
		return;
	Memoize_Frames (Cdr (thiso), last);
	Memoize_Frame (Car (thiso));
}

void Memoize_Frame (ELKObject frame) {
	ELKObject binding;

	for (; !Nullp (frame); frame = Cdr (frame)) {
		binding = Car (frame);
		SYMBOL(Car (binding))->value = Cdr (binding);
	}
}

void Forget_Frame (ELKObject frame) {
	for (; !Nullp (frame); frame = Cdr (frame))
		SYMBOL(Car (Car (frame)))->value = Unbound;
}

ELKObject Add_Binding (ELKObject frame, ELKObject sym, ELKObject val) {
	ELKObject b;
	GC_Node;

	GC_Link (frame);
	b = Cons (sym, val);
	GC_Unlink;
	return Cons (b, frame);
}

ELKObject Lookup_Symbol (ELKObject sym, int err) {
	ELKObject p, f, b;

	Env_To_List (The_Environment, p);
	for (; !Nullp (p); p = Cdr (p)) {
		for (f = Car (p); !Nullp (f); f = Cdr (f)) {   /* Inlined Assq() */
			b = Car (f);
			if (EQ(Car (b), sym))
				return b;
		}
	}
	if (err)
		Primitive_Error (x_("unbound variable: ~s"), sym);
	return Null;
}

ELKObject P_The_Environment (void) { return The_Environment; }

ELKObject P_Global_Environment (void) { return Global_Environment; }

ELKObject Define_Procedure (ELKObject form, ELKObject body, ELKObject sym) {
	ELKObject ret;
	GC_Node3;

	GC_Link3 (form, body, sym);
	body = Cons (Cdr (form), body);
	body = Cons (sym, body);
	body = Cons (body, Null);
	body = Cons (Car (form), body);
	ret = General_Define (body, sym);
	GC_Unlink;
	return ret;
}

ELKObject General_Define (ELKObject argl, ELKObject sym) {
	ELKObject val, var, frame, binding;
	GC_Node3;
	TC_Prolog;

	var = Car (argl);
	val = Cdr (argl);
	if (TYPE(var) == T_Symbol) {
		frame = Null;
		GC_Link3 (var, val, frame);
		if (Nullp (val)) {
			val = Void;
		} else {
			TC_Disable;
			val = Eval (Car (val));
			TC_Enable;
		}
		Set_Name (var, val);
		frame = Car (The_Environment);
		binding = Assq (var, frame);
		if (EQ(binding, Falseob)) {
			frame = Add_Binding (frame, var, val);
			Car (The_Environment) = frame;
		} else
			Cdr (binding) = val;
		SYMBOL(var)->value = val;
		GC_Unlink;
		return var;
	} else if (TYPE(var) == T_Pair) {
		if (Nullp (val))
			Primitive_Error (x_("no sub-forms in compound: ~s"), var);
		return Define_Procedure (var, val, sym);
	} else Wrong_Type_Combination (var, x_("symbol or pair"));
	/*NOTREACHED*/
	return(Null);
}

ELKObject P_Define (ELKObject argl) {
	return General_Define (argl, Sym_Lambda);
}

ELKObject P_Define_Macro (ELKObject argl) {
	return General_Define (argl, Sym_Macro);
}

ELKObject P_Set (ELKObject argl) {
	ELKObject val, var, binding, old;
	GC_Node3;
	TC_Prolog;

	var = Car (argl);
	val = Car (Cdr (argl));
	Check_Type (var, T_Symbol);
	binding = Lookup_Symbol (var, 1);
	old = Cdr (binding);
	GC_Link3 (var, binding, old);
	TC_Disable;
	val = Eval (val);
	TC_Enable;
	Set_Name (var, val);
	Cdr (binding) = val;
	SYMBOL(var)->value = val;
	GC_Unlink;
	return old;
}

void Set_Name (ELKObject var, ELKObject val) {
	register int t;

	t = TYPE(val);
	if (t == T_Compound) {
		if (Nullp (COMPOUND(val)->name))
			COMPOUND(val)->name = var;
	} else if (t == T_Macro) {
		if (Nullp (MACRO(val)->name))
			MACRO(val)->name = var;
	}
}

ELKObject P_Boundp (ELKObject x) {
	Check_Type (x, T_Symbol);
	return Nullp (Lookup_Symbol (x, 0)) ? Falseob : Trueob;
}
