/* -*- tab-width: 4 -*- */

/* Generic math functions.
 */

#include "config.h"
#include <math.h>
#include <errno.h>
#include <limits.h>
#ifndef MACOS
#  include <sys/types.h>
#endif
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include "elkkernel.h"

#if defined(MACOS) && !defined(MACOSX)
#  include <OSUtils.h>
#endif

#ifndef WIN32
extern int errno;
#else
#  include <process.h>
#endif

ELKObject Fixnum_To_String (ELKObject x, int radix);
ELKObject Generic_Multiply (ELKObject x, ELKObject y);
ELKObject Generic_Divide (ELKObject x, ELKObject y);
int Fixnum_Add (int a, int b, int *fits);
int Fixnum_Sub (int a, int b, int *fits);
ELKObject Fixnum_Multiply (int a, int b);
ELKObject General_Compare (int argc, ELKObject *argv, register int (*op)(ELKObject,ELKObject));
ELKObject Generic_Plus(ELKObject a, ELKObject b);
ELKObject Generic_Minus(ELKObject a, ELKObject b);
ELKObject General_Operator (int argc, ELKObject *argv, ELKObject start, ELKObject (*op)(ELKObject,ELKObject));
ELKObject General_Integer_Divide (ELKObject x, ELKObject y, int rem);
ELKObject Exact_Quotient(ELKObject x, ELKObject y);
ELKObject Exact_Remainder(ELKObject x, ELKObject y);
ELKObject Exact_Modulo(ELKObject x, ELKObject y);
ELKObject With_Exact_Ints (ELKObject x, ELKObject y, ELKObject (*fun)(ELKObject,ELKObject));
ELKObject Exact_Gcd(ELKObject x, ELKObject y);
ELKObject General_Gcd(ELKObject x, ELKObject y);
ELKObject Exact_Lcm(ELKObject x, ELKObject y);
ELKObject General_Lcm(ELKObject x, ELKObject y);
ELKObject General_Function (ELKObject x, ELKObject y, double (*fun)(double));
ELKObject Min (ELKObject x, ELKObject y);
ELKObject Max (ELKObject x, ELKObject y);

void Init_Math (void) {
#if defined(MACOS) && !defined(MACOSX)
	unsigned long rseed;
	GetDateTime(&rseed);
	srand(rseed);
#else
#  ifdef RANDOM
	srandom (getpid());
#  else
	srand (getpid());
#  endif
#endif
}

ELKObject Make_Integer (register int n) {
	ELKObject num;

	SET(num, T_Fixnum, n);
	return num;
}

ELKObject Make_Unsigned (register unsigned int n) {
	if (UFIXNUM_FITS(n))
		return Make_Integer (n);
	else
		return Unsigned_To_Bignum (n);
}

ELKObject Make_Long (register long n) {
	if (n < 0 ? (n < (long)INT_MIN) : (n > (long)INT_MAX))
		return Long_To_Bignum (n);
	else
		return Make_Integer ((int)n);
}

ELKObject Make_Unsigned_Long (register unsigned long n) {
	if ((n & ~((unsigned long)SIGNBIT-1)) == 0)
		return Make_Integer ((int)n);
	else
		return Unsigned_Long_To_Bignum (n);
}

ELKObject Fixnum_To_String (ELKObject x, int radix) {
	CHAR buf[32];
	register CHAR *p;
	register int n = FIXNUM(x), neg = 0;

	if (n == 0)
		return Make_String (x_("0"), 1);
	if (n < 0) {
		neg++;
		n = -n;
	}
	p = buf+31;
	*p = '\0';
	while (n > 0) {
		*--p = '0' + n % radix;
		if (*p > '9')
			*p = 'A' + (*p - '9') - 1;
		n /= radix;
	}
	if (neg)
		*--p = '-';
	return Make_String (p, estrlen (p));
}

CHAR *Flonum_To_String (ELKObject x) {
	static CHAR buf[32];
	CHAR *p;

	esnprintf (buf, 32, x_("%.15g"), FLONUM(x)->val);
	for (p = buf; *p; p++)
		if (*p == '.' || *p == 'e' || *p == 'N' || *p == 'i')
			return buf;
	*p++ = '.', *p++ = '0', *p++ = '\0';
	return buf;
}

ELKObject P_Number_To_String (int argc, ELKObject *argv) {
	int radix = 10;
	ELKObject x;
	CHAR *s;

	x = argv[0];
	if (argc == 2) {
		radix = Get_Exact_Integer (argv[1]);
		switch (radix) {
		case 2: case 8: case 10: case 16:
			break;
		default:
			Primitive_Error (x_("invalid radix: ~s"), argv[1]);
		}
	}
	Check_Number (x);
	switch (TYPE(x)) {
	case T_Fixnum:
		return Fixnum_To_String (x, radix);
	case T_Bignum:
		return Bignum_To_String (x, radix);
	case T_Flonum:
		if (radix != 10)
			Primitive_Error (x_("radix for reals must be 10"));   /* bleah! */
		s = Flonum_To_String (x);
		return Make_String (s, estrlen (s));
	}
	/*NOTREACHED*/
	return(Null);
}

int Get_Integer (ELKObject x) {
	double d;
	int expo;

	switch (TYPE(x)) {
	case T_Fixnum:
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Integer (x);
	case T_Flonum:
		d = FLONUM(x)->val;
		if (d != floor (d))
			Wrong_Type (x, T_Fixnum);
		(void)frexp (d, &expo);
		if (expo <= 8 * sizeof(int) - 1)
			return (int)d;
		Primitive_Error (x_("integer out of range: ~s"), x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

unsigned Get_Unsigned (ELKObject x) {
	double d;
	int expo;

	switch (TYPE(x)) {
	case T_Fixnum:
		if (FIXNUM(x) < 0)
			goto err;
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Unsigned (x);
	case T_Flonum:
		d = FLONUM(x)->val;
		if (d < 0)
			goto err;
		if (d != floor (d))
			Wrong_Type (x, T_Fixnum);
		(void)frexp (d, &expo);
		if (expo <= 8 * sizeof(int))
			return (unsigned)d;
err:
		Primitive_Error (x_("integer out of range: ~s"), x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

long Get_Long (ELKObject x) {
	double d;
	int expo;

	switch (TYPE(x)) {
	case T_Fixnum:
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Long (x);
	case T_Flonum:
		d = FLONUM(x)->val;
		if (d != floor (d))
			Wrong_Type (x, T_Fixnum);
		(void)frexp (d, &expo);
		if (expo <= 8 * sizeof(long) - 1)
			return (long)d;
		Primitive_Error (x_("integer out of range: ~s"), x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

unsigned long Get_Unsigned_Long (ELKObject x) {
	double d;
	int expo;

	switch (TYPE(x)) {
	case T_Fixnum:
		if (FIXNUM(x) < 0)
			goto err;
		return (unsigned long)FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Unsigned_Long (x);
	case T_Flonum:
		d = FLONUM(x)->val;
		if (d < 0)
			goto err;
		if (d != floor (d))
			Wrong_Type (x, T_Fixnum);
		(void)frexp (d, &expo);
		if (expo <= 8 * sizeof(long))
			return (unsigned long)d;
err:
		Primitive_Error (x_("integer out of range: ~s"), x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

int Get_Exact_Integer (ELKObject x) {
	switch (TYPE(x)) {
	case T_Fixnum:
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Integer (x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

unsigned Get_Exact_Unsigned (ELKObject x) {
	switch (TYPE(x)) {
	case T_Fixnum:
		if (FIXNUM(x) < 0)
			Primitive_Error (x_("integer out of range: ~s"), x);
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Unsigned (x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

long Get_Exact_Long (ELKObject x) {
	switch (TYPE(x)) {
	case T_Fixnum:
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Long (x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

unsigned long Get_Exact_Unsigned_Long (ELKObject x) {
	switch (TYPE(x)) {
	case T_Fixnum:
		if (FIXNUM(x) < 0)
			Primitive_Error (x_("integer out of range: ~s"), x);
		return FIXNUM(x);
	case T_Bignum:
		return Bignum_To_Unsigned_Long (x);
	default:
		Wrong_Type (x, T_Fixnum);
	}
	/*NOTREACHED*/
	return(0);
}

int Get_Index (ELKObject n, ELKObject obj) {
	register int size, i;

	i = Get_Exact_Integer (n);
	size = TYPE(obj) == T_Vector ? VECTOR(obj)->size : STRING(obj)->size;
	if (i < 0 || i >= size)
		Range_Error (n);
	return i;
}

ELKObject Make_Flonum (double d) {
	ELKObject num;

	num = Alloc_Object (sizeof (struct S_Flonum), T_Flonum, 0);
	FLONUM(num)->tag = Null;
	FLONUM(num)->val = d;
	return num;
}

ELKObject Make_Reduced_Flonum (double d) {
	ELKObject num;
	int expo;

	if (floor (d) == d) {
		if (d == 0)
			return Zero;
		(void)frexp (d, &expo);
		if (expo <= FIXBITS-1)
			return Make_Integer ((int)d);
	}
	num = Alloc_Object (sizeof (struct S_Flonum), T_Flonum, 0);
	FLONUM(num)->tag = Null;
	FLONUM(num)->val = d;
	return num;
}

int Fixnum_Add (int a, int b, int *fits) {
	int ret = a + b;

	*fits = 1;
	if (a > 0 && b > 0) {
		if (ret < 0) *fits = 0;
	} else if (a < 0 && b < 0) {
		if (ret > 0) *fits = 0;
	}
	return ret;
}

int Fixnum_Sub (int a, int b, int *fits) {
	int ret = a - b;

	*fits = 1;
	if (a < 0 && b > 0) {
		if (ret > 0) *fits = 0;
	} else if (a > 0 && b < 0) {
		if (ret < 0) *fits = 0;
	}
	return ret;
}

/* This function assumes 32bit integers.  This doesn't really matter,
 * because if the `*' primitive resorts to bignum multiplication, the
 * resulting bignum gets reduced to a fixnum (if it fits) anyway.
 * (This should be fixed, though...)
 */
ELKObject Fixnum_Multiply (int a, int b) {
	register unsigned aa = a;
	register unsigned ab = b;
	register unsigned prod, prod2;
	register int sign = 1;
	if (a < 0) {
		aa = -a;
		sign = -1;
	}
	if (b < 0) {
		ab = -b;
		sign = -sign;
	}
	prod = (aa & 0xFFFF) * (ab & 0xFFFF);
	if (aa & 0xFFFF0000) {
		if (ab & 0xFFFF0000)
			return Null;
		prod2 = (aa >> 16) * ab;
	} else {
		prod2 = aa * (ab >> 16);
	}
	prod2 += prod >> 16;
	prod &= 0xFFFF;
	if (prod2 > (1 << (FIXBITS - 1 - 16)) - 1) {
		if (sign == 1 || prod2 != (1 << (FIXBITS - 1 - 16)) || prod != 0)
			return Null;
#if 0		/* compiler hack */
		/*
		 * The SIGNBIT is a value with 1 in the sign bit and zeros elsewhere.
		 * Negating it implies reversing the bits and adding 1, which gives
		 * the same value.  However, some compilers notice that it is an overflow.
		 * Therefore, complementing the bits gives the proper intent: the largest
		 * integer.
		 *     -smr
		 */
		return Make_Integer (-(int)SIGNBIT);
#else
		return Make_Integer (~SIGNBIT);
#endif
	}
	prod += prod2 << 16;
	if (sign == -1)
		prod = (unsigned)(- (int)prod);
	return Make_Integer (prod);
}

ELKObject P_Integerp (ELKObject x) {
	double d;

	switch (TYPE(x)) {
	case T_Fixnum: case T_Bignum:
		return Trueob;
	case T_Flonum:
		d = FLONUM(x)->val;
		return d == floor(d) ? Trueob : Falseob;
	}
	return Falseob;
}

ELKObject P_Rationalp (ELKObject x) {
	return P_Integerp (x);
}

ELKObject P_Realp (ELKObject x) {
	register int t = TYPE(x);
	return t == T_Flonum || t == T_Fixnum  || t == T_Bignum ? Trueob : Falseob;
}

ELKObject P_Complexp (ELKObject x) {
	return P_Realp (x);
}

ELKObject P_Numberp (ELKObject x) {
	return P_Complexp (x);
}

ELKObject P_Exactp (ELKObject n) {
	Check_Number (n);
	return TYPE(n) == T_Flonum ? Falseob : Trueob;
}

ELKObject P_Inexactp (ELKObject n) {
	Check_Number (n);
	return TYPE(n) == T_Flonum ? Trueob : Falseob;
}

ELKObject P_Exact_To_Inexact (ELKObject n) {
	Check_Number (n);
	switch (TYPE(n)) {
	case T_Fixnum:
		return Make_Flonum ((double)FIXNUM(n));
	case T_Flonum:
		return n;
	case T_Bignum:
		return Make_Flonum (Bignum_To_Double (n));
	}
	/*NOTREACHED*/
	return(Null);
}

ELKObject P_Inexact_To_Exact (ELKObject n) {
	double d;
	int i;

	Check_Number (n);
	switch (TYPE(n)) {
	case T_Fixnum:
	case T_Bignum:
		return n;
	case T_Flonum:
		d = floor (FLONUM(n)->val + 0.5);
		(void)frexp (d, &i);
		return (i <= FIXBITS-1) ? Make_Integer ((int)d) : Double_To_Bignum (d);
	}
	/*NOTREACHED*/
	return(Null);
}

#define General_Generic_Predicate(prim,op,bigop) ELKObject prim (ELKObject x) {\
	register int ret;\
	Check_Number (x);\
	switch (TYPE(x)) {\
	case T_Flonum:\
		ret = FLONUM(x)->val op 0; break;\
	case T_Fixnum:\
		ret = FIXNUM(x) op 0; break;\
	case T_Bignum:\
		ret = bigop (x); break;\
	}\
	return ret ? Trueob : Falseob;\
}

General_Generic_Predicate (P_Zerop, ==, Bignum_Zero)
General_Generic_Predicate (P_Negativep, <, Bignum_Negative)
General_Generic_Predicate (P_Positivep, >, Bignum_Positive)

ELKObject P_Evenp (ELKObject x) {
	register int ret;
	double d;

	switch (TYPE(x)) {
	case T_Fixnum:
		ret = !(FIXNUM(x) & 1); break;
	case T_Bignum:
		ret = Bignum_Even (x); break;
	case T_Flonum:
		d = FLONUM(x)->val;
		if (floor (d) == d) {
			d /= 2;
			ret = floor (d) == d;
			break;
		}
		/*FALLTHROUGH*/
	default:
		Wrong_Type (x, T_Fixnum);
		/*NOTREACHED*/
	}
	return ret ? Trueob : Falseob;
}

ELKObject P_Oddp (ELKObject x) {
	ELKObject tmp;
	tmp = P_Evenp (x);
	return EQ(tmp,Trueob) ? Falseob : Trueob;
}

#define General_Generic_Compare(name,op,bigop) int name (ELKObject x, ELKObject y) {\
	ELKObject b; register int ret;\
	GC_Node;\
	\
	switch (TYPE(x)) {\
	case T_Fixnum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			return FIXNUM(x) op FIXNUM(y);\
		case T_Flonum:\
			return FIXNUM(x) op FLONUM(y)->val;\
		case T_Bignum:\
			GC_Link (y);\
			b = Integer_To_Bignum (FIXNUM(x));\
			ret = bigop (b, y);\
			GC_Unlink;\
			return ret;\
		}\
	case T_Flonum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			return FLONUM(x)->val op FIXNUM(y);\
		case T_Flonum:\
			return FLONUM(x)->val op FLONUM(y)->val;\
		case T_Bignum:\
			return FLONUM(x)->val op Bignum_To_Double (y);\
		}\
	case T_Bignum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			GC_Link (x);\
			b = Integer_To_Bignum (FIXNUM(y));\
			ret = bigop (x, b);\
			GC_Unlink;\
			return ret;\
		case T_Flonum:\
			return Bignum_To_Double (x) op FLONUM(y)->val;\
		case T_Bignum:\
			return bigop (x, y);\
		}\
	}\
	/*NOTREACHED*/ /* ...but lint never sees it */\
	return(0);\
}

General_Generic_Compare (Generic_Equal,	  ==, Bignum_Equal)
General_Generic_Compare (Generic_Less,		<, Bignum_Less)
General_Generic_Compare (Generic_Greater,	 >, Bignum_Greater)
General_Generic_Compare (Generic_Eq_Less,	<=, Bignum_Eq_Less)
General_Generic_Compare (Generic_Eq_Greater, >=, Bignum_Eq_Greater)

ELKObject General_Compare (int argc, ELKObject *argv, register int (*op)(ELKObject,ELKObject)) {
	register int i;

	Check_Number (argv[0]);
	for (i = 1; i < argc; i++) {
		Check_Number (argv[i]);
		if (!(*op) (argv[i-1], argv[i]))
			return Falseob;
	}
	return Trueob;
}

ELKObject P_Generic_Equal (int argc, ELKObject *argv) {
	return General_Compare (argc, argv, Generic_Equal);
}

ELKObject P_Generic_Less (int argc, ELKObject *argv) {
	return General_Compare (argc, argv, Generic_Less);
}

ELKObject P_Generic_Greater (int argc, ELKObject *argv) {
	return General_Compare (argc, argv, Generic_Greater);
}

ELKObject P_Generic_Eq_Less (int argc, ELKObject *argv) {
	return General_Compare (argc, argv, Generic_Eq_Less);
}

ELKObject P_Generic_Eq_Greater (int argc, ELKObject *argv) {
	return General_Compare (argc, argv, Generic_Eq_Greater);
}

#define General_Generic_Operator(name,op,fixop,bigop) ELKObject name (ELKObject x, ELKObject y) {\
	ELKObject b1, b2, ret; register int i;\
	int fits;\
	GC_Node2;\
	\
	switch (TYPE(x)) {\
	case T_Fixnum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			i = fixop (FIXNUM(x), FIXNUM(y), &fits);\
			if (fits)\
				return Make_Integer (i);\
			b1 = b2 = Null;\
			GC_Link2 (b1, b2);\
			b1 = Integer_To_Bignum (FIXNUM(x));\
			b2 = Integer_To_Bignum (FIXNUM(y));\
			ret = bigop (b1, b2);\
			GC_Unlink;\
			return ret;\
		case T_Flonum:\
			return Make_Flonum (FIXNUM(x) op FLONUM(y)->val);\
		case T_Bignum:\
			GC_Link (y);\
			b1 = Integer_To_Bignum (FIXNUM(x));\
			ret = bigop (b1, y);\
			GC_Unlink;\
			return ret;\
		}\
	case T_Flonum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			return Make_Flonum (FLONUM(x)->val op FIXNUM(y));\
		case T_Flonum:\
			return Make_Flonum (FLONUM(x)->val op FLONUM(y)->val);\
		case T_Bignum:\
			return Make_Flonum (FLONUM(x)->val op Bignum_To_Double (y));\
		}\
	case T_Bignum:\
		switch (TYPE(y)) {\
		case T_Fixnum:\
			GC_Link (x);\
			b1 = Integer_To_Bignum (FIXNUM(y));\
			ret = bigop (x, b1);\
			GC_Unlink;\
			return ret;\
		case T_Flonum:\
			return Make_Flonum (Bignum_To_Double (x) op FLONUM(y)->val);\
		case T_Bignum:\
			return bigop (x, y);\
		}\
	}\
	/*NOTREACHED*/ /* ...but lint never sees it */\
	return(Null);\
}

General_Generic_Operator (Generic_Plus,  +, Fixnum_Add, Bignum_Plus)
General_Generic_Operator (Generic_Minus, -, Fixnum_Sub, Bignum_Minus)

ELKObject P_Inc (ELKObject x) {
	Check_Number (x);
	return Generic_Plus (x, One);
}

ELKObject P_Dec (ELKObject x) {
	Check_Number (x);
	return Generic_Minus (x, One);
}

ELKObject General_Operator (int argc, ELKObject *argv, ELKObject start, register ELKObject (*op)(ELKObject,ELKObject)) {
	register int i;
	ELKObject accum;

	if (argc > 0)
		Check_Number (argv[0]);
	accum = start;
	switch (argc) {
	case 0:
		break;
	case 1:
		accum = (*op) (accum, argv[0]); break;
	default:
		for (accum = argv[0], i = 1; i < argc; i++) {
			Check_Number (argv[i]);
			accum = (*op) (accum, argv[i]);
		}
	}
	return accum;
}

ELKObject P_Generic_Plus (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, Zero, Generic_Plus);
}

ELKObject P_Generic_Minus (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, Zero, Generic_Minus);
}

ELKObject P_Generic_Multiply (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, One, Generic_Multiply);
}

ELKObject P_Generic_Divide (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, One, Generic_Divide);
}

ELKObject Generic_Multiply (ELKObject x, ELKObject y) {
	ELKObject b, ret;

	switch (TYPE(x)) {
	case T_Fixnum:
		switch (TYPE(y)) {
		case T_Fixnum:
			ret = Fixnum_Multiply (FIXNUM(x), FIXNUM(y));
			if (Nullp (ret)) {
				b = Integer_To_Bignum (FIXNUM(x));
				return Bignum_Fixnum_Multiply (b, y);
			}
			return ret;
		case T_Flonum:
			return Make_Flonum (FIXNUM(x) * FLONUM(y)->val);
		case T_Bignum:
			return Bignum_Fixnum_Multiply (y, x);
		}
	case T_Flonum:
		switch (TYPE(y)) {
		case T_Fixnum:
			return Make_Flonum (FLONUM(x)->val * FIXNUM(y));
		case T_Flonum:
			return Make_Flonum (FLONUM(x)->val * FLONUM(y)->val);
		case T_Bignum:
			return Make_Flonum (FLONUM(x)->val * Bignum_To_Double (y));
		}
	case T_Bignum:
		switch (TYPE(y)) {
		case T_Fixnum:
			return Bignum_Fixnum_Multiply (x, y);
		case T_Flonum:
			return Make_Flonum (Bignum_To_Double (x) * FLONUM(y)->val);
		case T_Bignum:
			return Bignum_Multiply (x, y);
		}
	}
	/*NOTREACHED*/
	return(Null);
}

ELKObject Generic_Divide (ELKObject x, ELKObject y) {
	register int t = TYPE(y);
	ELKObject b, ret;
	GC_Node2;

	if (t == T_Fixnum ? FIXNUM(y) == 0 :
		(t == T_Flonum ? FLONUM(y) == 0 : Bignum_Zero (y)))
		Range_Error (y);
	switch (TYPE(x)) {
	case T_Fixnum:
		switch (t) {
		case T_Fixnum:
			return Make_Reduced_Flonum ((double)FIXNUM(x) / (double)FIXNUM(y));
		case T_Flonum:
			return Make_Flonum ((double)FIXNUM(x) / FLONUM(y)->val);
		case T_Bignum:
			GC_Link (y);
			b = Integer_To_Bignum (FIXNUM(x));
			ret = Bignum_Divide (b, y);
			GC_Unlink;
			if (EQ(Cdr (ret),Zero))
				return Car (ret);
			return Make_Reduced_Flonum ((double)FIXNUM(x)
				/ Bignum_To_Double (y));
		}
	case T_Flonum:
		switch (t) {
		case T_Fixnum:
			return Make_Flonum (FLONUM(x)->val / (double)FIXNUM(y));
		case T_Flonum:
			return Make_Flonum (FLONUM(x)->val / FLONUM(y)->val);
		case T_Bignum:
			return Make_Flonum (FLONUM(x)->val / Bignum_To_Double (y));
		}
	case T_Bignum:
		switch (t) {
		case T_Fixnum:
			GC_Link (x);
			ret = Bignum_Fixnum_Divide (x, y);
			GC_Unlink;
			if (EQ(Cdr (ret),Zero))
				return Car (ret);
			return Make_Reduced_Flonum (Bignum_To_Double (x)
				/ (double)FIXNUM(y));
		case T_Flonum:
			return Make_Flonum (Bignum_To_Double (x) / FLONUM(y)->val);
		case T_Bignum:
			GC_Link2 (x, y);
			ret = Bignum_Divide (x, y);
			GC_Unlink;
			if (EQ(Cdr (ret),Zero))
				return Car (ret);
			return Make_Reduced_Flonum (Bignum_To_Double (x)
				/ Bignum_To_Double (y));
		}
	}
	/*NOTREACHED*/
	return(Null);
}

ELKObject P_Abs (ELKObject x) {
	register int i;

	Check_Number (x);
	switch (TYPE(x)) {
	case T_Fixnum:
		i = FIXNUM(x);
		return i < 0 ? Make_Integer (-i) : x;
	case T_Flonum:
		return Make_Flonum (fabs (FLONUM(x)->val));
	case T_Bignum:
		return Bignum_Abs (x);
	}
	/*NOTREACHED*/
	return(Null);
}

ELKObject General_Integer_Divide (ELKObject x, ELKObject y, int rem) {
	register int fx = FIXNUM(x), fy = FIXNUM(y);
	ELKObject b, ret;
	GC_Node;

	if (TYPE(y) == T_Fixnum ? FIXNUM(y) == 0 : Bignum_Zero (y))
		Range_Error (y);
	switch (TYPE(x)) {
	case T_Fixnum:
		switch (TYPE(y)) {
		case T_Fixnum:
			return Make_Integer (rem ? (fx % fy) : (fx / fy));
		case T_Bignum:
			GC_Link (y);
			b = Integer_To_Bignum (fx);
			GC_Unlink;
			ret = Bignum_Divide (b, y);
done:
			return rem ? Cdr (ret) : Car (ret);
		}
	case T_Bignum:
		switch (TYPE(y)) {
		case T_Fixnum:
			ret = Bignum_Fixnum_Divide (x, y);
			goto done;
		case T_Bignum:
			ret = Bignum_Divide (x, y);
			goto done;
		}
	}
	/*NOTREACHED*/
	return(Null);
}

ELKObject Exact_Quotient (ELKObject x, ELKObject y) {
	return General_Integer_Divide (x, y, 0);
}

ELKObject Exact_Remainder (ELKObject x, ELKObject y) {
	return General_Integer_Divide (x, y, 1);
}

ELKObject Exact_Modulo (ELKObject x, ELKObject y) {
	ELKObject rem, xneg, yneg;
	GC_Node2;

	GC_Link2 (x, y);
	rem = General_Integer_Divide (x, y, 1);
	xneg = P_Negativep (x);
	yneg = P_Negativep (y);
	if (!EQ(xneg,yneg))
		rem = Generic_Plus (rem, y);
	GC_Unlink;
	return rem;
}

ELKObject With_Exact_Ints (ELKObject x, ELKObject y, ELKObject (*fun)(ELKObject,ELKObject)) {
	ELKObject i, ret;
	int inex = 0;
	GC_Node3;

	ret = Null;
	GC_Link3 (x, y, ret);
	i = P_Integerp (x);
	if (!EQ(i,Trueob))
		Wrong_Type (x, T_Fixnum);
	i = P_Integerp (y);
	if (!EQ(i,Trueob))
		Wrong_Type (y, T_Fixnum);
	if (TYPE(x) == T_Flonum) {
		x = P_Inexact_To_Exact (x); inex++;
	}
	if (TYPE(y) == T_Flonum) {
		y = P_Inexact_To_Exact (y); inex++;
	}
	ret = fun (x, y);
	if (inex)
		ret = P_Exact_To_Inexact (ret);
	GC_Unlink;
	return ret;
}

ELKObject P_Quotient (ELKObject x, ELKObject y) {
	return With_Exact_Ints (x, y, Exact_Quotient);
}

ELKObject P_Remainder (ELKObject x, ELKObject y) {
	return With_Exact_Ints (x, y, Exact_Remainder);
}

ELKObject P_Modulo (ELKObject x, ELKObject y) {
	return With_Exact_Ints (x, y, Exact_Modulo);
}

ELKObject Exact_Gcd (ELKObject x, ELKObject y) {
	ELKObject r, z;
	GC_Node2;

	GC_Link2 (x, y);
	while (1) {
		z = P_Zerop (x);
		if (EQ(z,Trueob)) {
			r = y;
			break;
		}
		z = P_Zerop (y);
		if (EQ(z,Trueob)) {
			r = x;
			break;
		}
		r = General_Integer_Divide (x, y, 1);
		x = y;
		y = r;
	}
	GC_Unlink;
	return r;
}

ELKObject General_Gcd (ELKObject x, ELKObject y) {
	return With_Exact_Ints (x, y, Exact_Gcd);
}

ELKObject P_Gcd (int argc, ELKObject *argv) {
	return P_Abs (General_Operator (argc, argv, Zero, General_Gcd));
}

ELKObject Exact_Lcm (ELKObject x, ELKObject y) {
	ELKObject ret, p, z;
	GC_Node3;

	ret = Null;
	GC_Link3 (x, y, ret);
	ret = Exact_Gcd (x, y);
	z = P_Zerop (ret);
	if (!EQ(z,Trueob)) {
		p = Generic_Multiply (x, y);
		ret = General_Integer_Divide (p, ret, 0);
	}
	GC_Unlink;
	return ret;
}

ELKObject General_Lcm (ELKObject x, ELKObject y) {
	return With_Exact_Ints (x, y, Exact_Lcm);
}

ELKObject P_Lcm (int argc, ELKObject *argv) {
	return P_Abs (General_Operator (argc, argv, One, General_Lcm));
}

#define General_Conversion(name,op) ELKObject name (ELKObject x) {\
	double d, i;\
\
	Check_Number (x);\
	if (TYPE(x) != T_Flonum)\
		return x;\
	d = FLONUM(x)->val;\
	(void)modf (op (d), &i);\
	return Make_Flonum (i);\
}

#define trunc(x) (x)

General_Conversion (P_Floor, floor)
General_Conversion (P_Ceiling, ceil)
General_Conversion (P_Truncate, trunc)

ELKObject P_Round (ELKObject x) {
	double d, y, f;
	ELKObject ret, isodd;

	Check_Number (x);
	if (TYPE(x) != T_Flonum)
		return x;
	d = FLONUM(x)->val;
	y = d + 0.5;
	f = floor (y);
	ret = Make_Flonum (f);
	if (y == f) {
		isodd = P_Oddp (ret);
		if (Truep (isodd))
			FLONUM(ret)->val--;
	}
	return ret;
}

double Get_Double (ELKObject x) {
	Check_Number (x);
	switch (TYPE(x)) {
	case T_Fixnum:
		return (double)FIXNUM(x);
	case T_Flonum:
		return FLONUM(x)->val;
	case T_Bignum:
		return Bignum_To_Double (x);
	}
	/*NOTREACHED*/
	return(0);
}

ELKObject General_Function (ELKObject x, ELKObject y, double (*fun)(double)) {
	double d, ret;
	double (*twodouble)(double,double);

	d = Get_Double (x);
	errno = 0;
	if (Nullp (y))
		ret = (*fun) (d);
	else
	{
		twodouble = (double(*)(double,double))fun;
		ret = (*twodouble) (d, Get_Double (y));
	}
	if (errno == ERANGE || errno == EDOM)
		Range_Error (x);
	return Make_Flonum (ret);
}

ELKObject P_Sqrt (ELKObject x) { return General_Function (x, Null, sqrt); }

ELKObject P_Exp (ELKObject x) { return General_Function (x, Null, exp); }

ELKObject P_Log (ELKObject x) { return General_Function (x, Null, log); }

ELKObject P_Sin (ELKObject x) { return General_Function (x, Null, sin); }

ELKObject P_Cos (ELKObject x) { return General_Function (x, Null, cos); }

ELKObject P_Tan (ELKObject x) { return General_Function (x, Null, tan); }

ELKObject P_Asin (ELKObject x) { return General_Function (x, Null, asin); }

ELKObject P_Acos (ELKObject x) { return General_Function (x, Null, acos); }

ELKObject P_Atan (int argc, ELKObject *argv) {
	register int a2 = argc == 2;
	return General_Function (argv[0], a2 ? argv[1] : Null, a2 ?
		(double(*)(double))atan2 : (double(*)(double))atan);
}

ELKObject Min (ELKObject x, ELKObject y) {
	ELKObject ret;

	ret = Generic_Less (x, y) ? x : y;
	if (TYPE(x) == T_Flonum || TYPE(y) == T_Flonum)
		ret = P_Exact_To_Inexact (ret);
	return ret;
}

ELKObject Max (ELKObject x, ELKObject y) {
	ELKObject ret;

	ret = Generic_Less (x, y) ? y : x;
	if (TYPE(x) == T_Flonum || TYPE(y) == T_Flonum)
		ret = P_Exact_To_Inexact (ret);
	return ret;
}

ELKObject P_Min (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, argv[0], Min);
}

ELKObject P_Max (int argc, ELKObject *argv) {
	return General_Operator (argc, argv, argv[0], Max);
}

ELKObject P_Random (void) {
#ifdef RANDOM
	extern long random();
	return Make_Long (random ());
#else
	return Make_Integer (rand ());
#endif
}

ELKObject P_Srandom (ELKObject x) {
#ifdef RANDOM
	srandom (Get_Unsigned (x));
#else
	srand (Get_Unsigned (x));
#endif
	return x;
}
