/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include <math.h>


/*
 * electrical parameters used for deriving capacitance info for charge
 * sharing.  Default values aren't for any particular process, but are
 * self-consistent.
 *	Area capacitances are all in pfarads/sq-micron units.
 *	Perimeter capacitances are all in pfarads/micron units.
 */
double  irsim_CM2A = .00000;	/* 2nd metal capacitance -- area */
double  irsim_CM2P = .00000;	/* 2nd metal capacitance -- perimeter */
double  irsim_CMA  = .00003;	/* 1st metal capacitance -- area */
double  irsim_CMP  = .00000;	/* 1st metal capacitance -- perimeter */
double  irsim_CPA  = .00004;	/* poly capacitance -- area */
double  irsim_CPP  = .00000;	/* poly capacitance -- perimeter */
double  irsim_CDA  = .00010;	/* n-diffusion capacitance -- area */
double  irsim_CDP  = .00060;	/* n-diffusion capacitance -- perimeter */
double  irsim_CPDA = .00010;	/* p-diffusion capacitance -- area */
double  irsim_CPDP = .00060;	/* p-diffusion capacitance -- perimeter */
double  irsim_CGA  = .00040;	/* gate capacitance -- area */

				/* the following are computed from above */
double	irsim_CTDW;				/* xtor diff-width capacitance -- perimeter */
double	irsim_CPTDW;		
double	irsim_CTDE;				/* xtor diff-extension cap. -- perimeter */
double	irsim_CPTDE;	
double	irsim_CTGA;				/* xtor gate capacitance -- area */

double  irsim_LAMBDA     = 2.5;		/* microns/lambda */
double  irsim_LAMBDA2    = 6.25;	/* LAMBDA**2 */
long	irsim_LAMBDACM   = 250;		/* centi-microns/lambda */
double  irsim_LOWTHRESH  = 0.3;		/* low voltage threshold, normalized units */
double  irsim_HIGHTHRESH = 0.8;		/* high voltage threshold,normalized units */
double  irsim_DIFFEXT    = 0;		/* width of source/drain diffusion */

int     irsim_config_flags;

#define	CM_M	(100.0)		/* centimicrons per micron */

/* values of irsim_config_flags */

#define	CNTPULLUP	0x2		/* set if capacitance from gate of pullup */
							/* should be included. */

#define	DIFFPERIM	0x4		/* set if diffusion perimeter does not */
							/* include sources/drains of transistors. */

#define	SUBPAREA	0x8		/* set if poly over xistor doesn't make a capacitor */

#define	DIFFEXTF	0x10	/* set if we should add capacitance due to */
							/* diffusion-extension of source/drain. */


typedef struct	/* table translating parameters to its associated names */
{
	CHAR      *name;
	int       flag;
	double    *dptr;
} pTable;

static	pTable  parms[] = 
{
	x_("capm2a"),	    0x0,       &irsim_CM2A,
	x_("capm2p"),	    0x0,       &irsim_CM2P,
	x_("capma"),	    0x0,       &irsim_CMA,
	x_("capmp"),	    0x0,       &irsim_CMP,
	x_("cappa"),	    0x0,       &irsim_CPA,
	x_("cappp"),	    0x0,       &irsim_CPP,
	x_("capda"),	    0x0,       &irsim_CDA,
	x_("capdp"),	    0x0,       &irsim_CDP,
	x_("cappda"),	    0x0,       &irsim_CPDA,
	x_("cappdp"),	    0x0,       &irsim_CPDP,
	x_("capga"),	    0x0,       &irsim_CGA,
	x_("lambda"),	    0x0,       &irsim_LAMBDA,
	x_("lowthresh"),    0x0,       &irsim_LOWTHRESH,
	x_("highthresh"),   0x0,       &irsim_HIGHTHRESH,
	x_("diffperim"),    DIFFPERIM, NULL,
	x_("cntpullup"),    CNTPULLUP, NULL,
	x_("subparea"),	    SUBPAREA,  NULL,
	x_("diffext"),	    DIFFEXTF,  &irsim_DIFFEXT,
	NULL,           0x0,       NULL
};


#define	LSIZE		500		/* max size of parameter file input line */
#define	MAXARGS		10		/* max number of arguments in line */

static	int     lineno;		/* current line number */
static	CHAR    *currfile;	/* current input file */
static	int     nerrs;		/* errors found in config file */
static	int     maxerr;
static	CHAR	*ttype_drop[NTTYPES];


/*
 * info on resistance vs. width and length are stored first sorted by
 * width, then by length.
 */
struct length
{
	struct length    *next;	/* next element with same width */
	long             l;		/* length of this channel in centimicrons */
	double           r;		/* equivalent resistance/square */
};

struct width
{
	struct width     *next;	/* next width */
	long             w;		/* width of this channel in centimicrons */
	struct length    *list;	/* list of length structures */
} *resistances[R_TYPES][NTTYPES];


/* linear interpolation, assume that x1 < x <= x2 */
#define	interp(x, x1, y1, x2, y2)  \
	((((double) (x - x1)) / ((double) (x2 - x1))) * (y2 - y1) + y1)


typedef struct ResEntry  *resptr;

typedef struct ResEntry
{
	resptr    r_next;
	Resists   r;
} ResEntry;

#define	RES_TAB_SIZE	67
static resptr    *irsim_res_htab[NTTYPES];
static resptr    irsim_freeResist;


/* prototypes for local routines */
static int ParseLine(CHAR *line, CHAR **args);
static double lresist(struct length *list, long l, double size);
static double wresist(struct width *list, long w, long l);
static void linsert(struct length **list, long l, double resist);
static void winsert(struct width **list, long w, long l, double resist);
static void insert(CHAR *type, CHAR *context, CHAR *w, CHAR *l, CHAR *r);


/*
 * Routine to free all memory allocated in this module.
 */
void irsim_freeconfigmemory(void)
{
	static int first = 1;
	resptr *rtab, r;
	int i, n, c, t;
	struct width *w;
	struct length *l;

	if (first != 0)
	{
		for(i=0; i<NTTYPES; i++) irsim_res_htab[i] = 0;
		irsim_config_flags = 0;
		first = 0;
		irsim_freeResist = 0;
		nerrs = 0;
		for(t = 0; t < NTTYPES; t++)
			for(c = 0; c < R_TYPES; c++)
				resistances[c][t] = 0;
		return;
	}

	/* deallocate */
	for(i=0; i<NTTYPES; i++)
	{
		rtab = irsim_res_htab[i];
		if (rtab != 0)
		{
			for(n = 0; n < RES_TAB_SIZE; n++)
			{
				while (rtab[n] != NULL)
				{
					r = rtab[n];
					rtab[n] = r->r_next;
					efree((CHAR *)r);
				}
			}
			efree((CHAR *)irsim_res_htab[i]);
			irsim_res_htab[i] = 0;
		}
	}
	while (irsim_freeResist != 0)
	{
		r = irsim_freeResist;
		irsim_freeResist = r->r_next;
		efree((CHAR *)r);
	}
	for(t = 0; t < NTTYPES; t++)
	{
		for(c = 0; c < R_TYPES; c++)
		{
			w = resistances[c][t];
			while (resistances[c][t] != 0)
			{
				w = resistances[c][t];
				resistances[c][t] = w->next;
				while (w->list != 0)
				{
					l = w->list;
					w->list = l->next;
					efree((CHAR *)l);
				}
				efree((CHAR *)w);
			}
		}
	}
}


static int ParseLine(CHAR *line, CHAR **args)
{
	register CHAR  c;
	int            ac = 0;

	for(; ;)
	{
		while((c = *line) <= ' ' && c != '\0')
			line++;
		if (c == '\0' || c == ';')
			break;
		*args++ = line;
		ac++;
		while((c = *line) > ' ' && c != ';')
			line++;
		*line = '\0';
		if (c == '\0' || c == ';')
			break;
		line++;
	}
	*line = '\0';
	*args = NULL;
	return(ac);
}


int irsim_config(CHAR *cname)
{
	register pTable  *p;
	FILE             *cfile;
	CHAR             line[LSIZE], *dummy;
	CHAR             *targv[MAXARGS];
	CHAR             tmpbuff[NTTYPES * 22];
	int              targc;

	for(targc = 0; targc < NTTYPES; targc++)
	{
		ttype_drop[targc] = &tmpbuff[targc * 22];
		(void)esnprintf(ttype_drop[targc], 22, x_("%s-with-drop"), irsim_ttype[targc]);
	}

	currfile = cname;
	lineno = 0;
	cfile = xopen(cname, el_filetypetext, 0, &dummy);
	if (cfile == NULL)
	{
		ttyputmsg(x_("can't open electrical parameters file <%s>"), cname);
		return(1);
	}

	*line = '\0';
	(void)xfgets(line, LSIZE, cfile);
	if (estrncmp(line, x_("; configuration file"), 20))
	{
		rewind(cfile);
		maxerr = 1;
	}
	else
		maxerr = 15;

	while(!xfgets(line, LSIZE, cfile))
	{
		lineno++;
		targc = ParseLine(line, targv);
		if (targc == 0) continue;
		if (namesame(x_("resistance"), targv[0]) == 0)
		{
			if (targc >= 6)
				insert(targv[1], targv[2], targv[3], targv[4], targv[5]);
			else
			{
				irsim_error(currfile, lineno, x_("syntax error in resistance spec"));
				nerrs++;
			}
			continue;
		} else
		{
			for(p = parms; p->name != NULL; p++)
			{
				if (namesame(p->name, targv[0]) == 0)
				{
					if (p->dptr != NULL)
						*(p->dptr) = eatof(targv[1]);
					if (p->flag != 0 && eatof(targv[1]) != 0.0)
						irsim_config_flags |= p->flag;
					break;
				}
			}
			if (p->name == NULL)
			{
				irsim_error(currfile, lineno,
					x_("unknown electrical parameter: (%s)"), targv[0]);
				nerrs++;
			}
		}
		if (nerrs >= maxerr)
		{
			if (maxerr == 1)
				ttyputmsg(x_("I think %s is not an electrical parameters file"), cname);
			else
				ttyputmsg(x_("Too many errors in '%s'"), cname);
			return(1);
		}
	}
	irsim_LAMBDA2 = irsim_LAMBDA * irsim_LAMBDA;
	irsim_LAMBDACM = (long)(irsim_LAMBDA * CM_M);
	irsim_CTGA = ((irsim_config_flags & SUBPAREA) ? (irsim_CGA - irsim_CPA) : irsim_CGA) / (CM_M * CM_M);
	switch(irsim_config_flags & (DIFFEXTF | DIFFPERIM))
	{
		case 0:
			irsim_CTDE = irsim_CTDW = 0.0; irsim_CPTDE = irsim_CPTDW = 0.0;	
			break;
			case DIFFPERIM:
			irsim_config_flags |= TDIFFCAP;
			irsim_CTDE = irsim_CPTDE = 0.0;
			irsim_CTDW = -(irsim_CDP / CM_M);
			irsim_CPTDW = -(irsim_CPDP / CM_M);
			break;
		case DIFFEXTF:
			irsim_config_flags |= TDIFFCAP;
			irsim_CTDE = (2 * irsim_DIFFEXT * irsim_LAMBDA * irsim_CDP);
			irsim_CPTDE = (2 * irsim_DIFFEXT * irsim_LAMBDA * irsim_CPDP);
			irsim_CTDW = (irsim_CDP + irsim_DIFFEXT * irsim_LAMBDA * irsim_CDA) / CM_M;
			irsim_CPTDW = (irsim_CPDP + irsim_DIFFEXT * irsim_LAMBDA * irsim_CPDA) / CM_M;
			break;
		case (DIFFEXTF | DIFFPERIM):
			irsim_config_flags |= TDIFFCAP;
			irsim_CTDE = (2 * irsim_DIFFEXT * irsim_LAMBDA * irsim_CDP);
			irsim_CPTDE = (2 * irsim_DIFFEXT * irsim_LAMBDA * irsim_CPDP);
			irsim_CTDW = (irsim_DIFFEXT * irsim_LAMBDA * irsim_CDA) / CM_M;
			irsim_CPTDW = (irsim_DIFFEXT * irsim_LAMBDA * irsim_CPDA) / CM_M;
			break;
	}

	if (irsim_config_flags & CNTPULLUP)
		ttyputmsg(x_("warning: cntpullup is not supported"));

	(void)fclose(cfile);
	return(0);
}


/*
 * given a list of length structures, sorted by incresing length return
 * resistance of given channel.  If no exact match, return result of
 * linear interpolation using two closest channels.
 */
static double lresist(struct length *list, long l, double size)
{
	register struct length  *p, *q;

	for(p = list, q = NULL; p != NULL; q = p, p = p->next)
	{
		if (p->l == l ||(p->l > l && q == NULL))
			return(p->r * size);
		if (p->l > l)
			return(size * interp(l, q->l, q->r, p->l, p->r));
	}
	if (q != NULL)
		return(q->r *size);
	return(1E4 * size);
}


/*
 * given a pointer to the width structures for a particular type of
 * channel compute the resistance for the specified channel.
 */
static double wresist(struct width *list, long w, long l)
{
	register struct width  *p, *q;
	double                 size = ((double) l) / ((double) w);
	double                 temp;

	for(p = list, q = NULL; p != NULL; q = p, p = p->next)
	{
		if (p->w == w ||(p->w > w && q == NULL))
			return(lresist(p->list, l, size));
		if (p->w > w)
		{
			temp = lresist(q->list, l, size);
			return(interp(w, q->w, temp, p->w, lresist(p->list, l, size)));
		}
	}
	if (q != NULL)
		return(lresist(q->list, l, size));
	return(1E4 * size);
}


/*
 * Compute equivalent resistance given width, length and type of transistor.
 * for all contexts (STATIC, DYNHIGH, DYNLOW).  Place the result on the
 * transistor 
 */
Resists *irsim_requiv(int type, long width, long length)
{
	resptr           *rtab;
	register resptr  r;
	unsigned         n;

	type = BASETYPE(type);

	rtab = irsim_res_htab[type];
	if (rtab == NULL)
	{
		rtab = (resptr *)emalloc(RES_TAB_SIZE * sizeof(resptr *), sim_tool->cluster);
		for(n = 0; n < RES_TAB_SIZE; rtab[n++] = NULL);
		irsim_res_htab[type] = rtab;
	}
	n = ((unsigned) (length * 110133 + width)) % RES_TAB_SIZE;
	for(r = rtab[n]; r != NULL; r = r->r_next)
	{
		if ((long)r->r.length == length && (long)r->r.width == width) return(&r->r);
	}

	r = irsim_freeResist;
	if (r == NULL)
		r = (resptr)irsim_MallocList(sizeof(ResEntry), 1);
	irsim_freeResist = r->r_next;
	r->r_next = rtab[n];
	rtab[n] = r;

	r->r.length = length;
	r->r.width = width;

	if (type == RESIST)
	{
		r->r.dynlow = r->r.dynhigh = r->r.rstatic = (float) length / irsim_LAMBDACM;
	} else
	{
		r->r.rstatic = (float)wresist(resistances[STATIC][type], width, length);
		r->r.dynlow = (float)wresist(resistances[DYNLOW][type], width, length);
		r->r.dynhigh = (float)wresist(resistances[DYNHIGH][type], width, length);
	}
	return(&r->r);
}


static void linsert(struct length **list, long l, double resist)
{
	register struct length  *p, *q, *lnew;

	for(p = *list, q = NULL; p != NULL; q = p, p = p->next)
	{
		if (p->l == l)
		{
			p->r = resist;
			return;
		}
		if (p->l > l) break;
	}
	lnew = (struct length *)emalloc(sizeof(struct length), sim_tool->cluster);
	lnew->next = p;
	lnew->l = l;
	lnew->r = resist;
	if (q == NULL)
		*list = lnew;
	else
		q->next = lnew;
}


/* add a new data point to the interpolation array */
static void winsert(struct width **list, long w, long l, double resist)
{
	register struct width   *p, *q, *wnew;
	register struct length  *lnew;

	for(p = *list, q = NULL; p != NULL; q = p, p = p->next)
	{
		if (p->w == w)
		{
			linsert(&p->list, l, resist);
			return;
		}
		if (p->w > w) break;
	}
	wnew = (struct width *)emalloc(sizeof(struct width), sim_tool->cluster);
	lnew = (struct length *)emalloc(sizeof(struct length), sim_tool->cluster);
	wnew->next = p;
	wnew->list = lnew;
	wnew->w = w;
	if (q == NULL)
		*list = wnew;
	else
		q->next = wnew;
	lnew->next = NULL;
	lnew->l = l;
	lnew->r = resist;
}


/* interpret resistance specification command */
static void insert(CHAR *type, CHAR *context, CHAR *w, CHAR *l, CHAR *r)
{
	register int  c, t;
	long          width, length;
	double        resist;

	width = (long)(eatof(w) * CM_M);
	length = (long)(eatof(l) * CM_M);
	resist = eatof(r);
	if (width <= 0 || length <= 0 || resist <= 0)
	{
		irsim_error(currfile, lineno, x_("bad w, l, or r in config file"));
		nerrs++;
		return;
	}

	if (namesame(context, x_("static")) == 0)
		c = STATIC;
	else if (namesame(context, x_("dynamic-high")) == 0)
		c = DYNHIGH;
	else if (namesame(context, x_("dynamic-low")) == 0)
		c = DYNLOW;
	else if (namesame(context, x_("power")) == 0)
		c = POWER;
	else
	{
		irsim_error(currfile, lineno, x_("bad resistance context in config file"));
		nerrs++;
		return;
	}

	for(t = 0; t < NTTYPES; t++)
	{
		if (namesame(irsim_ttype[t], type) == 0)
		{
			if (c == POWER)
				return;
			winsert(&resistances[c][t], width, length, resist*width/length);
			return;
		}
		else if (namesame(ttype_drop[t], type) == 0)
			return;
	}

	irsim_error(currfile, lineno, x_("bad resistance transistor type"));
	nerrs++;
}
