/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */
#include "global.h"
#include "sim.h"
#include "irsimnet.h"
#include "irsimglobals.h"
#include "irsimnet_macros.h"


#define	DEFAULT_PERCENT	20		/* default % of nodes to seed */

#define	N_SEED0		0x080000
#define	N_SEED1		0x100000
#define	N_OBVIOUS0	0x200000
#define	N_OBVIOUS1	0x400000
#define	FAULT_ME	0x800000


#define	PRIM_OUTP	DELETED
#define	TRIGGER_EV	(CHNG_MODEL + 1)


typedef struct Trigger *ptrig;

typedef struct Trigger
{
	ptrig  next;	/* next trigger */
	TimeType period;	/* for sampler */
	TimeType offset;
	TimeType maxtime;
	hptr   samples;	/* list of sampling transitions */
	iptr   outp;	/* list of primary outputs to sample */
} Trigger;


static struct
{
	int    maybe;		/* may be detectable (node differs by X) */
	nptr   node;		/* node in which fault is detected */
	TimeType time;		/* time at which it is detected */
} fault;


static	FILE	*f_fault;			/* file to send output */
static	ptrig	triggers = NULL;		/* list of triggers */
static	TimeType maxSampleTime;
static	tptr    stuckTrans = NULL;		/* trans. to fault a node */
static	int	pend_triggers;			/* # pending trigger events */
static	int	n_nodes_seeded;			/* number of nodes seeded */
static	int	num_detect, num_maybe, num_fail;
#ifdef TIME_FAULTS
static	CHAR	ustr[40];
#endif

	/* forward declaration */
static tptr	init_fault_trans();

void irsim_init_faultsim_module(void)
{
	triggers = NULL;
	stuckTrans = NULL;
}

int irsim_add_trigger(nptr nd, int edge, TimeType delay)
{
	ptrig          t;
	struct Node    fake_nd;
	TimeType       tm;
	int            other = (edge == LOW) ? HIGH : LOW;
	register hptr  h;

	while(nd->nflags & ALIAS)
		nd = nd->nlink;
	fake_nd.curr = &fake_nd.head;
	fake_nd.head.next = irsim_last_hist;
	for(h = &nd->head; h != irsim_last_hist; h = h->next)
	{
		while(h != irsim_last_hist && (h->punt || (int)h->val != other)) h = h->next;
		while(h != irsim_last_hist && (h->punt || (int)h->val != edge)) h = h->next;
		tm = (TimeType)h->htime + delay;
		if ((int)h->val == edge && tm > 0 && tm <= irsim_cur_delta)
			irsim_AddHist(&fake_nd, edge, 1, tm, 0L, 0L);
	}

	if (fake_nd.head.next == irsim_last_hist)
		return(1);

	t = (ptrig)emalloc((int) sizeof(Trigger), sim_tool->cluster);
	t->samples = fake_nd.head.next;
	t->period = t->offset = 0;
	t->outp = NULL;
	t->next = triggers;
	triggers = t;
	return(0);
}


int irsim_add_sampler(TimeType period, TimeType offset)
{
	ptrig  t;

	maxSampleTime = irsim_cur_delta;
	t = (ptrig)emalloc((int) sizeof(Trigger), sim_tool->cluster);
	t->samples = NULL;
	t->period = period;
	t->offset = offset;
	t->outp = NULL;
	t->next = triggers;
	triggers = t;
	return(0);
}


int irsim_add_prim_output(nptr nd, int *flag)
{
	while(nd->nflags & ALIAS)
	nd = nd->nlink;

	if (nd->nflags & (POWER_RAIL | MERGED))
		return(1);
	if (nd->nflags & PRIM_OUTP)
	{
		if (nd->n.thev != (Thev) triggers)
		{
			ttyputmsg(x_("Too many trigger/sample for %s"), pnode(nd));
			*flag |= 2;
		}
	}
	else if (triggers != NULL)
	{
		irsim_iinsert(nd, &triggers->outp);
		nd->nflags |= PRIM_OUTP;
		nd->n.thev = (Thev) triggers;
		*flag |= 1;
	}
	return(1);
}


void irsim_cleanup_fsim(void)
{
	ptrig        t;
	iptr         i, it;
	struct Node  fake_nd;

	while(triggers != NULL)
	{
		if (triggers->samples != NULL)
		{
			fake_nd.head.next = triggers->samples;
			irsim_FreeHistList(&fake_nd);
		}
		for(i = triggers->outp; i != NULL; FreeInput(it))
		{
			it = i;
			i->inode->nflags &= ~PRIM_OUTP;
			i = i->next;
		}
		t = triggers;
		triggers = triggers->next;
		efree((CHAR *)t);
	}
}


#ifndef TIME_FAULTS
#  define FaultFail(ND, VAL) \
	{ num_fail ++; \
    if (f_fault != NULL) \
		efprintf(f_fault, x_("Fail\t%c  %s\n"), irsim_vchars[VAL], pnode(ND)); }

#  define PrFault(TYPE, N, V) \
    if (f_fault != NULL) \
      efprintf(f_fault, x_("%s\t%c  %s [%.2f] %s\n"), \
      TYPE, irsim_vchars[V], pnode(N), d2ns(fault.time), pnode(fault.node))

#else

#  define FaultFail(ND, VAL) \
    { num_fail ++; \
    if (f_fault != NULL) \
		efprintf(f_fault, x_("Fail\t%c  (%s) %s\n"), irsim_vchars[VAL], ustr, pnode(ND)); }

#  define PrFault(TYPE, N, V) \
    if (f_fault != NULL) \
      efprintf(f_fault, x_("%s\t%c  %s [%.2f] (%s) %s\n"), \
      TYPE, irsim_vchars[V], pnode(N), d2ns(fault.time), ustr, pnode(fault.node))

#endif

static void StuckNode(nptr nd, int val)
{
	tptr  t;

	if (stuckTrans == NULL) stuckTrans = init_fault_trans();

	t = stuckTrans;
	if (val == LOW)
	{
		t->ttype = NCHAN;	t->gate = irsim_VDD_node;	t->source = irsim_GND_node;
	}
	else	/* val == HIGH */
	{
		t->ttype = PCHAN;	t->gate = irsim_GND_node;	t->source = irsim_VDD_node;
	}

	t->drain = nd;
	CONNECT(t->gate->ngate, t);
	CONNECT(nd->nterm, t);
}


static void UnStuckNode(nptr nd)
{
	tptr  t = stuckTrans;

	DISCONNECT(t->gate->ngate, t);
	DISCONNECT(nd->nterm, t);
}


static int do_fault(nptr nd, int val)
{
	fault.maybe = FALSE;
	fault.node = NULL;

	StuckNode(nd, val);

#ifdef TIME_FAULTS
	uset_usage();    irsim_faultsim(nd);    get_usage(ustr);
#else
	irsim_faultsim(nd);
#endif

	if (fault.node != NULL)
	{
		num_detect++; PrFault(x_("Detect"), nd, val);
	} else if (fault.maybe)
	{
		num_maybe++; PrFault(x_("Maybe"), nd, val);
	} else
	{
		FaultFail(nd, val);
	}

	UnStuckNode(nd);

	return 0;
}


static int should_seed(nptr nd)
{
	register hptr  h;
	register int   zero, one;

	zero = one = 1;
	for(h = nd->head.next; h != irsim_last_hist; h = h->next)
	{
		if (h->inp)		return(0);	/* primary input */
		if (h->punt)		continue;
		if (h->val != LOW)	zero = 0;
		if (h->val != HIGH)	one = 0;
		if (zero + one == 0)	break;
	}

	nd->nflags |= (zero) ? N_OBVIOUS0 : N_SEED0;
	nd->nflags |= (one) ? N_OBVIOUS1 : N_SEED1;

	return(1);
}


static int seed_faults(int p_seed)
{
	nptr         n, nd;
	int          can_seed, num_to_seed, num;

	ttyputmsg(x_("seeding faults..."));

	nd = irsim_GetNodeList();

	can_seed = 0;
	for(n = nd; n != NULL; n = n->n.next)
	{
		if ((n->nflags & (ALIAS | MERGED | POWER_RAIL | INPUT)) == 0 &&
			n->ngate != NULL)
				can_seed += should_seed(n);
	}

	num_to_seed = (irsim_nnodes * p_seed) / 100;

	if (can_seed < num_to_seed)		/* seed all possible nodes */
	{
		for(n = nd; n != NULL; n = n->n.next)
		{
			if (n->nflags & (N_OBVIOUS0 | N_SEED0 | N_OBVIOUS1 | N_SEED1))
				n->nflags |= FAULT_ME;
		}
		ttyputmsg(x_("...done.  Only %d nodes to simulate"), can_seed);
		return(can_seed);
	}

	num = 0;
	srand((long) irsim_nnodes);

keep_seeding:
	for(n = nd; n != NULL; n = n->n.next)
	{
		if ((n->nflags & (N_OBVIOUS0 | N_SEED0 | N_OBVIOUS1 | N_SEED1)) == 0)
			continue;
		if (n->nflags & FAULT_ME)
			continue;
		if ((rand() & 100) <= p_seed)
		{
			num += 1;
			n->nflags |= FAULT_ME;
		}
		if (num >= num_to_seed)
			break;
	}

	if (num < num_to_seed)
	{
		p_seed += (p_seed + 1) / 2;	/* increase probability */
		goto keep_seeding;
	}

	ttyputmsg(x_("...done.  %d nodes to simulate"), can_seed);
	return(num_to_seed);
}


static int seed_fault(nptr nd)
{
	if (nd->nflags & FAULT_ME)
	{
		n_nodes_seeded ++;

		if (nd->nflags & N_OBVIOUS0)	FaultFail(nd, LOW);
		if (nd->nflags & N_SEED0)	do_fault(nd, LOW);

		if (nd->nflags & N_OBVIOUS1)	FaultFail(nd, HIGH);
		if (nd->nflags & N_SEED1)	do_fault(nd, HIGH);
	}
	nd->nflags &= ~(N_SEED0 | N_SEED1 | N_OBVIOUS0 | N_OBVIOUS1 | FAULT_ME);
	return(stopping(STOPREASONSIMULATE));
}


static int clear_node_flags(nptr nd)
{
	nd->nflags &= ~(N_SEED0 | N_SEED1 | N_OBVIOUS0 | N_OBVIOUS1 | FAULT_ME);
	return(0);
}


void irsim_exec_fsim(CHAR *fname, int p_seed)
{
	int     nodes_to_test;
	double  total;

	if (triggers == NULL)
	{
		ttyputmsg(x_("No triggers defined.  Aborted"));
		return;
	}

	if (p_seed <= 0) p_seed = DEFAULT_PERCENT;

	if (fname == NULL)
		fname = x_("fsim.out");
	if (estrcmp(fname, x_("/dev/null")) == 0)	/* for timing: no i/o */
		f_fault = NULL;
	else
	{
		f_fault = xcreate(fname, el_filetypetext, 0, 0);
		if (f_fault == NULL)
		{
			ttyputmsg(x_("Cannot open '%s'.  Aborted"));
			return;
		}
	}

	nodes_to_test = seed_faults(p_seed);

	num_detect = num_maybe = num_fail = n_nodes_seeded = 0;

	irsim_init_faultsim();

	irsim_walk_net(seed_fault);

	if (stopping(STOPREASONSIMULATE))
	{
		ttyputmsg(x_("** Interrupted ** => seeded %d nodes of %d (%g%%)"),
			n_nodes_seeded, nodes_to_test,
				100.0 * n_nodes_seeded / nodes_to_test);
		irsim_walk_net(clear_node_flags);
	}

	total = num_detect + num_maybe + num_fail;

	ttyputmsg(x_("----%.0f faults, %d detected (%d probably), %d undetected"),
		total, num_detect, num_maybe, num_fail);

	if (total == 0.0)
	{
		total = 1;
		num_detect = 1;
	}
	total *= 0.01;

	ttyputmsg(x_("fault coverage: %.2f%% (%.2f%%)"),
		num_detect / total, (num_detect + num_maybe) / total);

	if (f_fault != NULL && f_fault != stdout)
		(void)fclose(f_fault);

	irsim_end_faultsim();
}


void irsim_setup_triggers(void)
{
	ptrig  t;
	evptr  ev;

	pend_triggers = 0;
	for(t = triggers; t != NULL; t = t->next)
	{
		if (t->samples != NULL)
		{
			ev = irsim_EnqueueOther(TRIGGER_EV, (TimeType)t->samples->htime);
			ev->enode = (nptr) t;
			ev->p.hist = t->samples;
		} else
		{
			TimeType tm = t->offset != 0 ? t->offset : t->offset + t->period;
			ev = irsim_EnqueueOther(TRIGGER_EV, tm);
			ev->enode = (nptr) t;
		}
		pend_triggers ++;
	}
}


int irsim_do_trigger(evptr ev)
{
	ptrig          t = (ptrig) ev->enode;
	hptr           h;
	register iptr  i;

	for(i = t->outp; i != NULL; i = i->next)
	{
		register nptr  n = i->inode;

		if (n->nflags & DEVIATED)
		{
			if (n->npot != X && n->oldpot != X)
			{
				fault.time = irsim_cur_delta;
				fault.node = n;
				return(TRUE);
			}
			else if (! fault.maybe)
			{
				fault.time = irsim_cur_delta;
				fault.node = n;
				fault.maybe = TRUE;
			}
		}
	}
	if (t->samples != NULL)
	{
		h = ev->p.hist->next;
		if (h != irsim_last_hist)
		{
			ev = irsim_EnqueueOther(TRIGGER_EV, (TimeType)h->htime);
			ev->enode = (nptr) t;
			ev->p.hist = h;
		}
		else
			pend_triggers --;
	} else
	{
		if (ev->ntime + t->period <= maxSampleTime)
		{
			ev = irsim_EnqueueOther(TRIGGER_EV, ev->ntime + t->period);
			ev->enode = (nptr) t;
		}
		else
			pend_triggers --;
	}

	return((pend_triggers <= 0) ? TRUE : FALSE);
}


static tptr init_fault_trans(void)
{
	static Resists       smallR = { 0.00001f, 0.00001f, 0.00001f, 0, 0 };
	static struct Trans  tran;

	tran.scache.r = tran.dcache.r = NULL;
	tran.tflags = 0;
	tran.n_par = 0;
	tran.state = ON;
	tran.r = &smallR;
	tran.tlink = &tran;
	tran.x.pos = tran.x.pos = 0;
	return(&tran);
}
