/* -*- tab-width: 4 -*-
 *
 *     ********************************************************************* 
 *     * Copyright (C) 1988, 1990 Stanford University.                     * 
 *     * Permission to use, copy, modify, and distribute this              * 
 *     * software and its documentation for any purpose and without        * 
 *     * fee is hereby granted, provided that the above copyright          * 
 *     * notice appear in all copies.  Stanford University                 * 
 *     * makes no representations about the suitability of this            * 
 *     * software for any purpose.  It is provided "as is" without         * 
 *     * express or implied warranty.                                      * 
 *     ********************************************************************* 
 */

#include "global.h"
#include "irsimnet.h"
#include "irsimglobals.h"


#define	TSIZE		1024	/* size of event array, must be power of 2 */
#define	TMASK		(TSIZE - 1)

typedef struct
{
	evptr    flink, blink;	/* pointers in doubly-linked list */
} evhdr;


int    irsim_debug = 0;	    /* if <>0 print lotsa interesting info */

TimeType   irsim_cur_delta;	    /* current simulated time */
nptr   irsim_cur_node;	    /* node that belongs to current event */
long   irsim_nevent;		    /* number of current event */

evptr  irsim_evfree = NULL;	    /* list of free event structures */
int    irsim_npending;	    /* number of pending events */


static	evhdr  ev_array[TSIZE];	    /* used as head of doubly-linked lists */
#define EV_LIST(T) (evptr)(&ev_array[(Uint)(T) & TMASK])

void irsim_init_sched(void)
{
	irsim_debug = 0;
	irsim_evfree = NULL;
}

/*
 * Sets the parameter "list" to contain a pointer to the list of pending 
 * events at time "irsim_cur_delta + delta".  "end_of_list" contains a pointer to
 * the end of the list. If there are no events at the indicated time "list"
 * is set to NULL.  The function returns 0 if there are no events past the
 * "irsim_cur_delta + delta", otherwise it returns the delta at which the next
 * events are found;
 */

TimeType irsim_pending_events(TimeType delta, evptr *list, evptr *end_of_list)
{
	evhdr           *hdr;
	register evptr  ev;

	*list = NULL;

	delta += irsim_cur_delta;
	hdr = &ev_array[(Uint)delta & TMASK];

	if (hdr->flink == (evptr) hdr || hdr->blink->ntime < delta)
		goto find_next;

	for(ev = hdr->flink; ev->ntime < delta; ev = ev->flink);
		if (ev->ntime != delta)
			goto find_next;

	*list = ev;
	if (hdr->blink->ntime == ev->ntime)
		*end_of_list = hdr->blink;
	else
	{
		for(delta = ev->ntime; ev->ntime == delta; ev = ev->flink);
			*end_of_list = ev->blink;
	}

find_next:
	{
		TimeType i, limit, time;

		time = irsim_max_time;
		for(i = ++delta, limit = i + TSIZE; i < limit; i++)
		{
			ev = EV_LIST(i);
			if (ev != ev->flink)
			{
				if (ev->blink->ntime < delta)
					continue;
				for(ev = ev->flink; ev->ntime < delta; ev = ev->flink);
					if (ev->ntime < limit)
				{
					time = ev->ntime;
					break;
				}
				else if (ev->flink->ntime < time)
					time = ev->flink->ntime;
			}
		}
		delta = (time != irsim_max_time) ? time - irsim_cur_delta : 0;
	}

	return(delta);
}


/*
 * find the next event to be processed by scanning event wheel.  Return
 * the list of events to be processed at this time, removing it first
 * from the time wheel.
 */
evptr irsim_get_next_event(TimeType stop_time)
{
	register evptr  event;
	TimeType        i, limit, time;

	if (irsim_npending == 0)
		return(NULL);

	time = irsim_max_time;
	for(i = irsim_cur_delta, limit = i + TSIZE; i < limit; i++)
	{
		event = EV_LIST(i);
		if (event != event->flink)
		{
			if (event->flink->ntime < limit)		/* common case */
				goto found;
			if (event->flink->ntime < time)
				time = event->flink->ntime;
		}
	}

	if (time != (long)irsim_max_time)
		event = EV_LIST(time);
	else
	{
		ttyputmsg(x_("*** internal error: no events but npending set"));
		return(NULL);
	}

found:
	{
		evptr  evlist = event->flink;

		time = evlist->ntime;

		if (time >= stop_time)
			return(NULL);

		irsim_cur_delta = time;			/* advance simulation time */

		if (event->blink->ntime != time)	/* check tail of list */
		{
			do
				event = event->flink;
			while(event->ntime == time);

			event = event->blink;		/* grab part of the list */
			evlist->blink->flink = event->flink;
			event->flink->blink = evlist->blink;
			evlist->blink = event;
			event->flink = NULL;
		} else
		{
			event = evlist->blink;		/* grab the entire list */
			event->blink->flink = NULL;
			evlist->blink = event->blink;
			event->flink = event->blink = (evptr) event;
		}
		return(evlist);
	}
}


/*
 * remove event from all structures it belongs to and return it to free pool
 */
void irsim_free_event(register evptr event)
{
	/* unhook from doubly-linked event list */
	event->blink->flink = event->flink;
	event->flink->blink = event->blink;
	irsim_npending -= 1;

	/* add to free storage pool */
	event->flink = irsim_evfree;
	irsim_evfree = event;

	free_from_node(event, event->enode);
}

/* get event structure.  Allocate a bunch more if we've run out. */
#define NEW_EVENT(NEW)					\
{								\
	if (((NEW) = irsim_evfree) == NULL)				\
		(NEW) = (evptr) irsim_MallocList(sizeof(struct Event), 1);\
	irsim_evfree = (NEW)->flink;					\
}


/*
 * Add an event to event list, specifying transition delay and new value.
 * 0 delay transitions are converted into unit delay transitions (0.01 ns).
 */
void irsim_enqueue_event(register nptr n, int newvalue, long delta, long rtime)
{
	register evptr  marker, newev;
	TimeType   etime;

	NEW_EVENT(newev);

	/* remember facts about this event */
	newev->ntime = etime = irsim_cur_delta + (TimeType)delta;
	newev->rtime = (short)rtime;
	newev->enode = n;
	newev->p.cause = irsim_cur_node;
	newev->delay = delta;
	if (newvalue == DECAY)		/* change value to X here */
	{
		newev->eval = X;
		newev->type = DECAY_EV;
	} else
	{
		newev->eval = (UCHAR1)newvalue;
		newev->type = REVAL;		/* for incremental simulation */
	}

	/* add the new event to the event list at the appropriate entry
	* in event wheel.  Event lists are kept sorted by increasing
	* event time.
	*/
	marker = EV_LIST(etime);

	/* Check whether we need to insert-sort in the list */
	if ((marker->blink != marker) && (marker->blink->ntime > etime))
	{
		do { marker = marker->flink; } while (marker->ntime <= etime);
	}

	/* insert event right before event pointed to by marker */
	newev->flink = marker;
	newev->blink = marker->blink;
	marker->blink->flink = newev;
	marker->blink = newev;
	irsim_npending += 1;

	/* 
	 * thread event onto list of events for this node, keeping it
	 * in sorted order
	 */
	if ((n->events != NULL) && (n->events->ntime > etime))
	{
		for(marker = n->events; (marker->nlink != NULL) &&
			(marker->nlink->ntime > etime); marker = marker->nlink);
		newev->nlink = marker->nlink;
		marker->nlink = newev;
	} else
	{
		newev->nlink = n->events;
		n->events = newev;
	}
}


/* same as irsim_enqueue_event, but assumes 0 delay and rtise/fall time */
void irsim_enqueue_input(register nptr n, int newvalue)
{
	register evptr  marker, newev;
	TimeType   etime;

	/* Punt any pending events for this node. */
	while(n->events != NULL)
		irsim_free_event(n->events);

	NEW_EVENT(newev);

	/* remember facts about this event */
	newev->ntime = etime = irsim_cur_delta;
	newev->rtime = 0;   newev->delay = 0;
	newev->enode = newev->p.cause = n;
	newev->eval = (UCHAR1)newvalue;
	newev->type = REVAL;			/* anything, doesn't matter */

	/* Add new event to HEAD of list at appropriate entry in event wheel */

	marker = EV_LIST(etime);
	newev->flink = marker->flink;
	newev->blink = marker;
	marker->flink->blink = newev;
	marker->flink = newev;
	irsim_npending += 1;

	/* thread event onto (now empty) list of events for this node */
	newev->nlink = NULL;
	n->events = newev;
}


/*
 * Initialize event structures
 */
void irsim_init_event(void)
{
	register int    i;
	register evhdr  *event;

	for(i = 0, event = &ev_array[0]; i < TSIZE; i += 1, event += 1)
	{
		event->flink = event->blink = (evptr) event;
	}
	irsim_npending = 0;
	irsim_nevent = 0;
}


void irsim_PuntEvent(nptr node, evptr ev)
{
	if (node->nflags & WATCHED)
		ttyputmsg(x_("    punting transition of %s -> %c scheduled for %2.2fns"),
			node->nname, irsim_vchars[ev->eval], d2ns(ev->ntime));

	if (ev->type != DECAY_EV)		/* don't save punted decay events */
		irsim_AddPunted(ev->enode, ev, irsim_cur_delta);
	irsim_free_event(ev);
}


void irsim_requeue_events(evptr evlist, int thread)
{
	TimeType   etime;
	register evptr  ev, next, target;

	irsim_npending = 0;
	for(ev = evlist; ev != NULL; ev = next)
	{
		next = ev->flink;

		irsim_npending++;
		etime = ev->ntime;
		target = EV_LIST(etime);

		if ((target->blink != target) && (target->blink->ntime > etime))
		{
			do { target = target->flink; } while(target->ntime <= etime);
		}

		ev->flink = target;
		ev->blink = target->blink;
		target->blink->flink = ev;
		target->blink = ev;

		if (thread)
		{
			register nptr   n = ev->enode;
			register evptr  marker;

			if ((n->events != NULL) && (n->events->ntime > etime))
			{
				for(marker = n->events; (marker->nlink != NULL) &&
					(marker->nlink->ntime > etime); marker = marker->nlink);
				ev->nlink = marker->nlink;
				marker->nlink = ev;
			} else
			{
				ev->nlink = n->events;
				n->events = ev;
			}
		}
	}
}

	/* Incremental simulation routines */

/*
 * Back the event queues up to time 'btime'.  This is the opposite of
 * advancing the simulation time.  Mark all pending events as PENDING,
 * and re-enqueue them according to their creation-time (ntime - delay).
 */
evptr irsim_back_sim_time(TimeType btime, int is_inc)
{
	evptr           tmplist;
	register int    nevents;
	register evptr  ev, next;
	register evhdr  *hdr, *endhdr;

	nevents = 0;
	tmplist = NULL;

	/* first empty out the time wheel onto the temporary list */
	for(endhdr = &ev_array[TSIZE], hdr = ev_array; hdr != endhdr ; hdr++)
	{
		for(ev = hdr->flink; ev != (evptr) hdr; ev = next)
		{
			next = ev->flink;

			ev->blink->flink = ev->flink;	/* remove event */
			ev->flink->blink = ev->blink;
			if (is_inc)
				free_from_node(ev, ev->enode);

			if ((! is_inc) && ev->ntime - ev->delay >= btime)
			{
				free_from_node(ev, ev->enode);
				ev->flink = irsim_evfree;
				irsim_evfree = ev;
			} else
			{
				ev->flink = tmplist;		/* move it to tmp list */
				tmplist = ev;

				nevents++;
			}
		}
	}

	if (! is_inc)
	{
		irsim_requeue_events(tmplist, FALSE);
		return(NULL);
	}

	if (is_inc != TRUE)	/* only for fault simulation (is_inc == 2) */
	{
		irsim_npending = 0;
		return(tmplist);
	}

	/* now move the temporary list to the time wheel */
	for(ev = tmplist; ev != NULL; ev = next)
	{
		TimeType   etime;
		register evptr  target;

		next = ev->flink;

		ev->ntime -= (TimeType)ev->delay;
		ev->type = PENDING;
		etime = ev->ntime;
		target = EV_LIST(etime);

		if ((target->blink != target) && (target->blink->ntime > etime))
		{
			do { target = target->flink; } while(target->ntime <= etime);
		}

		ev->flink = target;
		ev->blink = target->blink;
		target->blink->flink = ev;
		target->blink = ev;
	}

	irsim_npending = nevents;
	return(NULL);
}


/*
 * Enqueue event type 'type' form history entry 'hist' for node 'nd'.
 * Note that events with type > THREAD are not threaded onto a node's list
 * of pending events, since we don't want the new_val routines to look at
 * this event, instead we keep track of these events in the c.event event
 * of every node.  Return FALSE if this history entry is the sentinel
 * (irsim_last_hist), otherwise return TRUE.
 */
int irsim_EnqueueHist(nptr nd, hptr hist, int type)
{
	register evptr  marker, newev;
	TimeType   etime;

	if (hist == irsim_last_hist)			/* never queue this up */
	{
		nd->c.event = NULL;
		return(FALSE);
	}

	NEW_EVENT(newev);

	/* remember facts about this event */
	newev->ntime = etime = (TimeType)hist->htime;
	newev->eval = hist->val;
	newev->enode = nd;
	newev->p.hist = hist;
	if (hist->punt)
	{
		newev->delay = hist->t.p.delay;
		newev->rtime = hist->t.p.rtime;
	} else
	{
		newev->delay = hist->t.r.delay;
		newev->rtime = hist->t.r.rtime;
	}

	marker = EV_LIST(etime);

	/* Check whether we need to insert-sort in the list */
	if ((marker->blink != marker) && (marker->blink->ntime > etime))
	{
		do { marker = marker->flink; } while(marker->ntime <= etime);
	}

	/* insert event right before event pointed to by marker */
	newev->flink = marker;
	newev->blink = marker->blink;
	marker->blink->flink = newev;
	marker->blink = newev;
	irsim_npending += 1;

	if (hist->inp)
		type |= IS_INPUT;
	else if (newev->delay == 0)
		type |= IS_XINPUT;
	newev->type = (UCHAR1)type;

	if (type > THREAD)
	{
		nd->c.event = newev;
		return(TRUE);
	}

	if ((nd->events != NULL) && (nd->events->ntime > etime))
	{
		for(marker = nd->events; (marker->nlink != NULL) &&
			(marker->nlink->ntime > etime); marker = marker->nlink);
		newev->nlink = marker->nlink;
		marker->nlink = newev;
	} else
	{
		newev->nlink = nd->events;
		nd->events = newev;
	}
	return(TRUE);
}


void irsim_DequeueEvent(register nptr nd)
{
	register evptr  ev;

	ev = nd->c.event;
	ev->blink->flink = ev->flink;
	ev->flink->blink = ev->blink;
	ev->flink = irsim_evfree;
	irsim_evfree = ev;
	nd->c.event = NULL;

	irsim_npending -= 1;
}


void irsim_DelayEvent(evptr ev, TimeType delay)
{
	register evptr  marker, newev;
	TimeType   etime;
	register nptr   nd;

	nd = ev->enode;
	NEW_EVENT(newev);

	/* remember facts about this event */
	*newev = *ev;
	newev->delay += (long)delay;
	newev->ntime += delay;

	etime = newev->ntime;

	marker = EV_LIST(etime);

	/* Check whether we need to insert-sort in the list */
	if ((marker->blink != marker) && (marker->blink->ntime > etime))
	{
		do { marker = marker->flink; } while(marker->ntime <= etime);
	}

	/* insert event right before event pointed to by marker */
	newev->flink = marker;
	newev->blink = marker->blink;
	marker->blink->flink = newev;
	marker->blink = newev;
	irsim_npending += 1;

	if (newev->type > THREAD)
	{
		nd->c.event = newev;
		return;
	}

	if ((nd->events != NULL) && (nd->events->ntime > etime))
	{
		for(marker = nd->events; (marker->nlink != NULL) &&
			(marker->nlink->ntime > etime); marker = marker->nlink);
		newev->nlink = marker->nlink;
		marker->nlink = newev;
	} else
	{
		newev->nlink = nd->events;
		nd->events = newev;
	}
}


evptr irsim_EnqueueOther(int type, TimeType time)
{
	register evptr  marker, newev;
	TimeType   etime;

	NEW_EVENT(newev);

	newev->ntime = etime = time;
	newev->type = (UCHAR1)type;

	marker = EV_LIST(etime);

	/* Check whether we need to insert-sort in the list */
	if ((marker->blink != marker) && (marker->blink->ntime > etime))
	{
		do { marker = marker->flink; } while(marker->ntime <= etime);
	}

	/* insert event right before event pointed to by marker */
	newev->flink = marker;
	newev->blink = marker->blink;
	marker->blink->flink = newev;
	marker->blink = newev;
	irsim_npending += 1;
	return(newev);
}


/*
 * Remove any events that may be left from the incremental simulation.
 */
void irsim_rm_inc_events(int all)
{
	register int    nevents;
	register evptr  ev, next;
	register evhdr  *hdr, *endhdr;

	nevents = 0;

	for(endhdr = &ev_array[TSIZE], hdr = ev_array; hdr != endhdr; hdr++)
	{
		for(ev = hdr->flink; ev != (evptr) hdr; ev = next)
		{
			next = ev->flink;
			if (all || ev->type >= THREAD)
			{
				ev->blink->flink = next;		/* remove event */
				ev->flink->blink = ev->blink;
				ev->flink = irsim_evfree;
				irsim_evfree = ev;
				if (ev->type <= THREAD)
					free_from_node(ev, ev->enode);
			}
			else
				nevents++;
		}
	}
	irsim_npending = nevents;
}
